unit Numedit;

// Copyright(C) 2012 Roland Hahn
// This library is free software; you can redistribute it and/or modify it.

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, LResources, Forms, Controls, Graphics, Dialogs, StdCtrls,
  Clipbrd;

type
  TNumEdit = class(TCustomEdit)
  private
    { Private-Deklarationen }
    FEntertaste: boolean;
    FWert: extended;
    FMatrix: string;
    FWarnton: boolean;
    procedure SetEntertaste(AEntertaste: boolean);
    procedure SetWert(AWert: extended);
    procedure SetMatrix(AMatrix: string);
    procedure SetWarnton(AWarnton: boolean);
    function GetKommastellen(): integer;
    function GetFormat(): string;
  protected
    { Protected-Deklarationen }
    procedure Click; override;
    procedure KeyPress(var Key: char); override;
    procedure Change; override;
    procedure DoExit; override;
    procedure DoEnter; override;
  public
    { Public-Deklarationen }
    constructor Create(AOwner: TComponent); override;
  published
    { Published-Deklarationen }
    property Alignment;
    property AutoSelect;
    property AutoSize;
    property BorderStyle;
    property Color;
    property Enabled;
    property Font;
    property Height;
    property HelpContext;
    property Hint;
    property Left;
    property Name;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ReadOnly;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Tag;
    property Text;
    property Top;
    property Visible;
    property Width;

    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;

    property Entertaste: boolean read FEntertaste write SetEntertaste default True;
    property Wert: extended read FWert write SetWert;
    property Matrix: string read FMatrix write SetMatrix;
    property Warnton: boolean read FWarnton write SetWarnton default True;

  end;

procedure Register;

implementation

var
  Zeichen: string;
  BackText: string;
  BackSelStart: integer;

//------------------------------------------------------------------------------

procedure Register;
begin
  RegisterComponents('rh Tools', [TNumedit]);
end;
//------------------------------------------------------------------------------

constructor TNumEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FEntertaste := True;
  FWarnton := True;
  {$IFDEF UNIX}
  DefaultFormatSettings.DecimalSeparator := ',';
  DefaultFormatSettings.ThousandSeparator := '.';
  {$ENDIF}
  FMatrix := '999999' + DefaultFormatSettings.DecimalSeparator + '99';
  Alignment := taRightJustify;
  Zeichen := #3 + #8 + #13 + #22 + #24 + DefaultFormatSettings.DecimalSeparator +
    '-0123456789';
end;
//------------------------------------------------------------------------------

function TNumedit.GetKommastellen(): integer;
begin
  Result := Pos(DefaultFormatSettings.DecimalSeparator, Matrix);
  if Result > 0 then
    Result := Length(Matrix) - Result;
end;
//------------------------------------------------------------------------------

function TNumedit.GetFormat(): string;
var
  i: integer;
begin
  Result := '0.';
  for i := 1 to GetKommastellen do
    Result := Result + '0';
end;
//------------------------------------------------------------------------------

procedure TNumEdit.SetEntertaste(AEntertaste: boolean);
begin
  FEntertaste := AEntertaste;
end;
//------------------------------------------------------------------------------

procedure TNumEdit.SetWarnton(AWarnton: boolean);
begin
  FWarnton := AWarnton;
end;
//------------------------------------------------------------------------------

procedure TNumedit.SetWert(AWert: extended);
begin
  if FWert <> AWert then
  begin
    FWert := AWert;
    if GetKommastellen > 0 then
      Text := FormatFloat(GetFormat, FWert)
    else
      Text := FloatToStr(Round(FWert));
  end;
end;
//------------------------------------------------------------------------------

procedure TNumEdit.SetMatrix(AMatrix: string);
begin
  // Sicherstellen, dass in Matrix immer der richtige Dezimaltrenner steht.
  AMatrix := StringReplace(AMatrix, '.', DefaultFormatSettings.DecimalSeparator,
    [rfReplaceAll]);
  AMatrix := StringReplace(AMatrix, ',', DefaultFormatSettings.DecimalSeparator,
    [rfReplaceAll]);
  FMatrix := AMatrix;
end;
//------------------------------------------------------------------------------

procedure TNumedit.KeyPress(var Key: char);
begin
  BackSelStart := SelStart;

  // Wenn eine unerlaubte Taste gedrückt wurde, Key löschen.
  if Pos(Key, Zeichen) = 0 then
  begin
    Key := #0;
    if Warnton then
      Beep();
  end;

  // Wenn der Wert bereits eine führende Null enthält, wird abgebrochen.
  if (Key = '0') and (SelLength = 0) then
  begin
    if (LeftStr(Text, 2) = '-0') and (SelStart < 3) then
      Key := #0;
    if (LeftStr(Text, 1) = '0') and (SelStart < 2) then
      Key := #0;
    if (Key = #0) and Warnton then
      Beep();
  end;

  // Wenn Entertaste gedrückt wurde und Enter = True ist, zur nächsten TAB-Position springen.
  if (Key = #13) and Entertaste then // #13 =Enter
  begin
    Key := #0; // Beep verhindern.
    Screen.ActiveForm.SelectNext(Screen.ActiveForm.ActiveControl, True, True);
  end;

  inherited KeyPress(Key); // übergibt an die geerbte KeyPress-Methode
end;
//------------------------------------------------------------------------------

procedure TNumedit.Change;
var
  Z: extended;
  OK: boolean;
begin
  if Text = '' then
  begin
    FWert := 0;
    BackText := '';
    Exit;
  end;

  if (Text = '-') and (Matrix[1] = '-') then
  begin
    FWert := 0;
    if BackText <> '-' then
      BackText := '-';
    BackSelStart := 1;
    Exit;
  end;

  if (Text = DefaultFormatSettings.DecimalSeparator) and (GetKommastellen > 0) then
  begin
    FWert := 0;
    if BackText <> DefaultFormatSettings.DecimalSeparator then
      BackText := DefaultFormatSettings.DecimalSeparator;
    BackSelStart := 1;
    Exit;
  end;

  if (Text = '-' + DefaultFormatSettings.DecimalSeparator) and
    ((Matrix[1] = '-') and (GetKommastellen > 0)) then
  begin
    FWert := 0;
    if BackText <> '-' + DefaultFormatSettings.DecimalSeparator then
      BackText := '-' + DefaultFormatSettings.DecimalSeparator;
    BackSelStart := 2;
    Exit;
  end;

  Z := 0;
  OK := True;

  // Wenn der Wert mehr Dezimalstellen enthält als Matrix, wird abgebrochen.
  if Pos(DefaultFormatSettings.DecimalSeparator, Text) > 0 then
  begin
    if GetKommastellen = 0 then
      OK := False;
    if (Length(Text) - Pos(DefaultFormatSettings.DecimalSeparator, Text)) >
      GetKommastellen then
      OK := False;
  end;

  // Wenn der Wert Negativ ist, Matrix aber keine Negativwerte vorsieht, wird abgebrochen.
  if (Matrix[1] <> '-') and (Pos('-', Text) > 0) then
    OK := False;

  // Wenn sich kein gültiger Wert ergibt, wird abgebrochen.
  if not TryStrToFloat(Text, Z) then
    OK := False;

  // Wenn der Absolutwert des Wertes größer ist, als der Absolutwert aus Matrix, wird abgebrochen.
  if Abs(Z) > Abs(StrToFloat(Matrix)) then
    OK := False;

  if OK then
  begin
    FWert := Z;
    if BackText <> Text then
      BackText := Text;
  end
  else
  begin
    Text := BackText;
    SelStart := BackSelStart;
    if Warnton then
      Beep();
    Exit;
  end;

  inherited Change; // übergibt an die geerbte Change-Methode
end;
//------------------------------------------------------------------------------

procedure TNumedit.Click;
begin
  BackSelStart := SelStart;

  inherited Click; // übergibt an die geerbte Click-Methode
end;
//------------------------------------------------------------------------------

procedure TNumedit.DoEnter;
begin
  BackText := Text;
  BackSelStart := SelStart;

  inherited DoEnter; // übergibt an die geerbte DoEnter-Methode
end;
//------------------------------------------------------------------------------

procedure TNumedit.DoExit;
begin
  if (Text <> '') and (Text <> '-') and
    (Text <> DefaultFormatSettings.DecimalSeparator) and
    (Text <> '-' + DefaultFormatSettings.DecimalSeparator) then
  begin
    if GetKommastellen > 0 then
      Text := FormatFloat(GetFormat, FWert)
    else
      Text := FloatToStr(Round(FWert));
  end
  else
  begin
    Text := '';
    FWert := 0;
  end;

  inherited DoExit; // übergibt an die geerbte DoExit-Methode
end;
//------------------------------------------------------------------------------

initialization
  {$I NumEdit.lrs}

end.
